# pyOCD debugger
# Copyright (c) 2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyocd.debug.svd.loader import (
    SVDFile,
    SVDLoader,
)

class TestIntervalSvdAccess:
    def builtin_svd(self, name: str) -> SVDLoader:
        def completion(dev):
            pass
        loader = SVDLoader(SVDFile.from_builtin(name), completion)
        loader.run()
        return loader

    def test_load(self):
        loader = self.builtin_svd('Musca_B1.svd')
        assert loader.device
        assert [p for p in loader.device.peripherals if p.name == 'UART0']

